#!/usr/bin/env python

"""
Inspect source files, obtaining details of classes and attributes.

Copyright (C) 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import compiler.ast
from compiler.visitor import ASTVisitor
try:
    set
except NameError:
    from sets import Set as set

class InspectError(Exception):

    "An inspection error."

    pass

class Class:

    "An inspected class."

    def __init__(self, name):
        self.name = name
        self.bases = []
        self.namespace = {}
        self.instattr = set()

    def add_base(self, base):
        self.bases.append(base)

class Function:

    "An inspected function."

    def __init__(self, name):
        self.name = name

class Module(ASTVisitor):

    "An inspected module."

    def __init__(self, importer=None):
        ASTVisitor.__init__(self)
        self.visitor = self

        self.importer = importer

        self.namespace = {}
        self.namespace.update(builtins_namespace)

        self.in_global = 1
        self.in_init = 0
        self.expr = None

        self.classes = []
        self.module = None

    def parse(self, filename):
        module = compiler.parseFile(filename)
        self.process(module)

    def process(self, module):
        self.module = module
        return self.dispatch(module)

    def default(self, node, *args):
        raise InspectError, (None, node)

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def store(self, name, obj):
        if self.in_global:
            if not self.classes:
                self.namespace[name] = obj
            else:
                self.classes[-1].namespace[name] = obj

    def store_attr(self, name):
        if self.in_init:
            self.classes[-1].instattr.add(name)

    def NOP(self, node):
        return node

    visitAnd = NOP

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssAttr(self, node):
        expr = self.dispatch(node.expr)
        if expr is not None and isinstance(expr, Self):
            self.store_attr(node.attrname)
        return None

    def visitAssList(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssName(self, node):
        self.store(node.name, self.expr)
        return None

    visitAssTuple = visitAssList

    visitCallFunc = NOP

    def visitClass(self, node):
        if not self.in_global:
            raise InspectError, "Class is %s not global: cannot handle this reasonably." % node.name
        else:
            cls = Class(node)
            for base in node.bases:
                base_ref = self.dispatch(base)
                if base_ref is None:
                    raise InspectError, "Base class %s for class %s is not found: it may be hidden in some way." % (base, node.name)
                cls.add_base(base_ref)

            # Make an entry for the class.

            self.store(node.name, cls)

            self.classes.append(cls)
            self.dispatch(node.code)
            self.classes.pop()

        return node

    visitConst = NOP

    visitDict = NOP

    visitDiscard = NOP

    visitFor = NOP

    def visitFrom(self, node):
        if self.importer is None:
            raise InspectError, "Please use the micropython.Importer class for code which uses the 'from' statement."

        module = self.importer.load(node.modname, 1)

        if module is None:
            print "Warning:", node.modname, "not imported."
            return None

        for name, alias in node.names:
            if name != "*":
                self.namespace[alias] = module.namespace[name]
            else:
                for n in module.namespace.keys():
                    self.namespace[n] = module.namespace[n]

        return None

    def visitFunction(self, node):
        self.store(node.name, Function(node.name))

        in_global = self.in_global
        self.in_global = 0
        if node.name == "__init__" and self.classes:
            self.in_init = 1
            self.dispatch(node.code)
            self.in_init = 0
        self.in_global = in_global
        return None

    def visitGetattr(self, node):
        expr = self.dispatch(node.expr)
        if expr is not None and isinstance(expr, Module):
            return expr.namespace.get(node.attrname)
        else:
            return None

    def visitIf(self, node):
        for test, body in node.tests:
            self.dispatch(body)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    def visitImport(self, node):
        if self.importer is None:
            raise InspectError, "Please use the micropython.Importer class for code which uses the 'import' statement."

        for name, alias in node.names:
            self.namespace[alias] = self.importer.load(name, 1)

        return None

    visitList = NOP

    def visitModule(self, node):
        return self.dispatch(node.node)

    def visitName(self, node):
        name = node.name
        if name == "self":
            return Self()
        elif self.namespace.has_key(name):
            return self.namespace[name]
        else:
            return None

    visitNot = NOP

    visitOr = NOP

    visitPass = NOP

    visitRaise = NOP

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    visitSubscript = NOP

    def visitTryExcept(self, node):
        self.dispatch(node.body)
        for name, var, n in node.handlers:
            self.dispatch(n)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

class Self:

    "A reference to an object within a method."

    pass

builtins_namespace = {}
for key in ['ArithmeticError', 'AssertionError', 'AttributeError',
    'BaseException', 'DeprecationWarning', 'EOFError', 'Ellipsis',
    'EnvironmentError', 'Exception', 'False', 'FloatingPointError',
    'FutureWarning', 'GeneratorExit', 'IOError', 'ImportError',
    'ImportWarning', 'IndentationError', 'IndexError', 'KeyError',
    'KeyboardInterrupt', 'LookupError', 'MemoryError', 'NameError',
    'None', 'NotImplemented', 'NotImplementedError', 'OSError',
    'OverflowError', 'PendingDeprecationWarning', 'ReferenceError',
    'RuntimeError', 'RuntimeWarning', 'StandardError', 'StopIteration',
    'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit',
    'TabError', 'True', 'TypeError', 'UnboundLocalError',
    'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError',
    'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning',
    'ValueError', 'Warning', 'ZeroDivisionError']:
    builtins_namespace[key] = Class(key)

# vim: tabstop=4 expandtab shiftwidth=4
